"""Provide builting schedulers"""
from __future__ import annotations

from typing import Type

from xqute import Scheduler
from xqute.schedulers.local_scheduler import (
    LocalJob as XquteLocalJob,
    LocalScheduler as XquteLocalScheduler,
)
from xqute.schedulers.sge_scheduler import (
    SgeJob as XquteSgeJob,
    SgeScheduler as XquteSgeScheduler
)
from xqute.schedulers.slurm_scheduler import (
    SlurmJob as XquteSlurmJob,
    SlurmScheduler as XquteSlurmScheduler,
)
from xqute.schedulers.ssh_scheduler import (
    SshJob as XquteSshJob,
    SshScheduler as XquteSshScheduler,
)

from .defaults import SCHEDULER_ENTRY_GROUP
from .exceptions import NoSuchSchedulerError, WrongSchedulerTypeError
from .job import Job
from .utils import is_subclass, load_entrypoints


class LocalJob(XquteLocalJob, Job):
    """Job class for local scheduler"""


class LocalScheduler(XquteLocalScheduler):
    """Local scheduler"""
    job_class = LocalJob


class SgeJob(XquteSgeJob, Job):
    """Job class for SGE scheduler"""


class SgeScheduler(XquteSgeScheduler):
    """SGE scheduler"""
    job_class = SgeJob


class SlurmJob(XquteSlurmJob, Job):
    """Job class for Slurm scheduler"""


class SlurmScheduler(XquteSlurmScheduler):
    """Slurm scheduler"""
    job_class = SlurmJob


class SshJob(XquteSshJob, Job):
    """Job class for SSH scheduler"""


class SshScheduler(XquteSshScheduler):
    """SSH scheduler"""
    job_class = SshJob


def get_scheduler(scheduler: str | Type[Scheduler]) -> Type[Scheduler]:
    """Get the scheduler by name of the scheduler class itself

    Args:
        scheduler: The scheduler class or name

    Returns:
        The scheduler class
    """
    if is_subclass(scheduler, Scheduler):
        return scheduler  # type: ignore

    if scheduler == "local":
        return LocalScheduler

    if scheduler == "sge":
        return SgeScheduler

    if scheduler == "slurm":
        return SlurmScheduler

    if scheduler == "ssh":
        return SshScheduler

    for n, obj in load_entrypoints(SCHEDULER_ENTRY_GROUP):  # pragma: no cover
        if n == scheduler:
            if not is_subclass(obj, Scheduler):
                raise WrongSchedulerTypeError(
                    "Scheduler should be a subclass of "
                    "pipen.scheduler.Scheduler."
                )
            return obj

    raise NoSuchSchedulerError(str(scheduler))



"""Provide the PipelinePBar and ProcPBar classes"""
from __future__ import annotations

from typing import TYPE_CHECKING

from .utils import truncate_text

if TYPE_CHECKING:  # pragma: no cover
    import enlighten

# [12/02/20 12:44:06] I core
#                 pipeline: 100%|
# |        desc_len      |
PBAR_DESC_LEN = 24


class ProcPBar:
    """The progress bar for processes"""

    def __init__(
        self, manager: enlighten.Manager, proc_size: int, proc_name: str
    ) -> None:
        self.submitted_counter = manager.counter(
            total=proc_size,
            color="cyan",
            desc=proc_name,
            unit="jobs",
            leave=False,
        )
        self.running_counter = self.submitted_counter.add_subcounter("yellow")
        self.success_counter = self.submitted_counter.add_subcounter("green")
        self.failure_counter = self.submitted_counter.add_subcounter("red")

    def update_job_submitted(self):
        """Update the progress bar when a job is submitted"""
        self.submitted_counter.update()

    def update_job_retrying(self):
        """Update the progress bar when a job is retrying"""
        # self.running_counter.count -= 1
        self.failure_counter.update(-1)

    def update_job_running(self):
        """Update the progress bar when a job is running"""
        try:
            self.running_counter.update_from(self.submitted_counter)
        except ValueError:  # pragma: no cover
            pass

    def update_job_succeeded(self):
        """Update the progress bar when a job is succeeded"""
        try:
            self.success_counter.update_from(self.running_counter)
        except ValueError:  # pragma: no cover
            self.success_counter.update_from(self.submitted_counter)
        except:  # noqa: E722  # pragma: no cover
            pass

    def update_job_failed(self):
        """Update the progress bar when a job is failed"""
        try:
            self.failure_counter.update_from(self.running_counter)
        except ValueError:  # pragma: no cover
            try:
                self.failure_counter.update_from(self.submitted_counter)
            except ValueError:  # pragma: no cover
                pass
        except:  # noqa: E722  # pragma: no cover
            pass

    def done(self):
        """The process is done"""
        self.submitted_counter.close()


class PipelinePBar:
    """Progress bar for the pipeline"""

    def __init__(self, n_procs: int, ppln_name: str) -> None:
        """Initialize progress bar for pipeline"""
        import enlighten

        desc_len = PBAR_DESC_LEN
        ppln_name = truncate_text(ppln_name, desc_len)
        self.manager = enlighten.get_manager()
        self.running_counter = self.manager.counter(
            total=n_procs,
            color="yellow",
            desc=f"{ppln_name:>{desc_len}}:",
            unit="procs",
        )
        self.success_counter = self.running_counter.add_subcounter("green")
        self.failure_counter = self.running_counter.add_subcounter("red")
        self.desc_len = desc_len

    def proc_bar(self, proc_size: int, proc_name: str) -> ProcPBar:
        """Get the progress bar for a process

        Args:
            proc_size: The size of the process
            proc_name: The name of the process

        Returns:
            The progress bar for the given process
        """
        proc_name = truncate_text(proc_name, self.desc_len)
        proc_name = f"{proc_name:>{self.desc_len}}:"
        return ProcPBar(self.manager, proc_size, proc_name)

    def update_proc_running(self):
        """Update the progress bar when a process is running"""
        self.running_counter.update()

    def update_proc_done(self):
        """Update the progress bar when a process is done"""
        self.success_counter.update_from(self.running_counter)

    def update_proc_error(self):
        """Update the progress bar when a process is errored"""
        self.failure_counter.update_from(self.running_counter)

    def done(self) -> None:
        """When the pipeline is done"""
        self.running_counter.close()
        self.manager.stop()



"""Main entry module, provide the Pipen class"""
from __future__ import annotations

import asyncio
from os import PathLike
from pathlib import Path
from typing import Any, ClassVar, Iterable, List, Sequence, Type

from diot import Diot
from rich import box
from rich.panel import Panel
from rich.text import Text
from simpleconf import ProfileConfig
from varname import varname, VarnameException

from .defaults import CONFIG, CONFIG_FILES
from .exceptions import (
    PipenOrProcNameError,
    ProcDependencyError,
    PipenSetDataError,
)
from .pluginmgr import plugin
from .proc import Proc
from .progressbar import PipelinePBar
from .utils import (
    copy_dict,
    desc_from_docstring,
    get_logpanel_width,
    get_plugin_context,
    is_valid_name,
    log_rich_renderable,
    logger,
    pipen_banner,
)


class Pipen:
    """The Pipen class provides interface to assemble and run the pipeline

    Attributes:
        name: The name of the pipeline
        desc: The description of the pipeline
        outdir: The output directory of the results
        procs: The processes
        pbar: The progress bar
        starts: The start processes
        config: The configurations
        workdir: The workdir for the pipeline
        profile: The profile of the configurations to run the pipeline
        _kwargs: The extra configrations passed to overwrite the default ones

        PIPELINE_COUNT: How many pipelines are loaded
        SETUP: Whether the one-time setup hook is called

    Args:
        name: The name of the pipeline
        desc: The description of the pipeline
        outdir: The output directory of the results
        **kwargs: Other configurations
    """

    PIPELINE_COUNT: ClassVar[int] = 0
    SETUP: ClassVar[bool] = False

    name: str = None
    desc: str = None
    outdir: str | PathLike = None
    starts: List[Proc] = []
    data: Iterable = None
    # other configs

    def __init__(
        self,
        name: str = None,
        desc: str = None,
        outdir: str | PathLike = None,
        **kwargs,
    ) -> None:
        """Constructor"""
        self.procs: List[Proc] = None
        self.pbar: PipelinePBar = None
        if name is not None:
            self.name = name
        elif self.__class__.name is not None:
            self.name = self.__class__.name
        else:
            try:
                self.name = varname()
            except VarnameException:
                if self.__class__.PIPELINE_COUNT == 0:
                    self.name = self.__class__.__name__
                else:
                    self.name = (
                        f"{self.__class__.__name__}-"
                        f"{self.__class__.PIPELINE_COUNT}"
                    )

        if not is_valid_name(self.name):
            raise PipenOrProcNameError(
                fr"Invalid pipeline name: {self.name}, expecting '^[\w.-]$'"
            )

        self.desc = (
            desc
            or self.__class__.desc
            or desc_from_docstring(self.__class__, Pipen)
        )
        self.outdir = Path(
            outdir or self.__class__.outdir or f"./{self.name}-output"
        ).resolve()
        self.workdir: Path = None
        self.profile: str = "default"

        self.starts: List[Proc] = self.__class__.starts
        if self.starts and not isinstance(self.starts, (tuple, list)):
            self.starts = [self.starts]

        self.config = Diot(copy_dict(CONFIG, 3))
        # We shouldn't update the config here, since we don't know
        # the profile yet
        self._kwargs = {
            key: value
            for key, value in self.__class__.__dict__.items()
            if key in self.config
        }
        self._kwargs.setdefault("plugin_opts", {}).update(
            kwargs.pop("plugin_opts", {})
        )
        self._kwargs.setdefault("template_opts", {}).update(
            kwargs.pop("template_opts", {})
        )
        self._kwargs.setdefault("scheduler_opts", {}).update(
            kwargs.pop("scheduler_opts", {})
        )
        self._kwargs.update(kwargs)
        # Initialize the workdir, as workdir is created before _init()
        # But the config is updated in _init()
        # Here we hack it to have the workdir passed in.
        if "workdir" in kwargs:
            self.config.workdir = kwargs["workdir"]

        if not self.__class__.SETUP:  # pragma: no cover
            # Load plugins from entrypotins at runtime to avoid
            # cyclic imports
            plugin.load_entrypoints()

        plugins = self.config.plugins or self._kwargs.get("plugins", [])
        self.plugin_context = get_plugin_context(plugins)
        self.plugin_context.__enter__()

        # make sure core plugin is enabled
        plugin.get_plugin("core").enable()

        if not self.__class__.SETUP:  # pragma: no cover
            plugin.hooks.on_setup(self.config)
            self.__class__.SETUP = True

        self.__class__.PIPELINE_COUNT += 1

        if self.__class__.data is not None:
            self.set_data(*self.__class__.data)

    def __init_subclass__(cls) -> None:
        cls.PIPELINE_COUNT = 0

    async def async_run(self, profile: str = "default") -> bool:
        """Run the processes one by one

        Args:
            profile: The default profile to use for the run

        Returns:
            True if the pipeline ends successfully else False
        """
        self.profile = profile
        self.workdir = Path(self.config.workdir) / self.name
        # self.workdir.mkdir(parents=True, exist_ok=True)

        succeeded = True
        await self._init()
        logger.setLevel(self.config.loglevel.upper())
        log_rich_renderable(pipen_banner(), "magenta", logger.info)
        try:
            self.build_proc_relationships()
            self._log_pipeline_info()
            await plugin.hooks.on_start(self)
            for proc in self.procs:
                self.pbar.update_proc_running()
                proc_obj = proc(self)  # type: ignore
                if proc in self.starts and proc.input_data is None:
                    proc_obj.log(
                        "warning",
                        "This is a start process, "
                        "but no 'input_data' specified.",
                    )
                await proc_obj.init()
                await proc_obj.run()
                if proc_obj.succeeded:
                    self.pbar.update_proc_done()
                else:
                    self.pbar.update_proc_error()
                    succeeded = False
                    break
                proc_obj.gc()

            logger.info("")
        except Exception:
            raise
        else:
            await plugin.hooks.on_complete(self, succeeded)
        finally:
            self.plugin_context.__exit__()
            if self.pbar:
                self.pbar.done()

        return succeeded

    def run(
        self,
        profile: str = "default",
    ) -> int:
        """Run the pipeline with the given profile
        This is just a sync wrapper for the async `async_run` function using
        `asyncio.run()`

        Args:
            profile: The default profile to use for the run

        Returns:
            True if the pipeline ends successfully else False
        """
        return asyncio.run(self.async_run(profile))

    def set_data(self, *indata: Any) -> Pipen:
        """Set the input_data for start processes

        Args:
            *indata: The input data for the start processes
                The data will set for the processes in the order determined by
                `set_starts()`.
                If a process has input_data set, an error will be raised.
                To use that input_data, set None here in the corresponding
                position for the process

        Raises:
            ProcInputDataError: When trying to set input data to
                processes with input_data already set

        Returns:
            `self` to chain the operations
        """
        for start, data in zip(self.starts, indata):
            if data is None:
                continue
            if start.input_data is not None:
                raise PipenSetDataError(
                    f"`input_data` has already set for {start}. "
                    "If you want to use it, set `None` at the position of "
                    "this process for `Pipen.set_data()`."
                )
            start.input_data = data
        return self

    def set_starts(
        self,
        *procs: Type[Proc] | Sequence[Type[Proc]],
        clear: bool = True,
    ):
        """Set the starts

        Args:
            *procs: The processes to set as starts of the pipeline.
            clear: Wether to clear previous set starts

        Raises:
            ProcDependencyError: When processes set as starts repeatedly

        Returns:
            `self` to chain the operations
        """
        if clear:
            self.starts = []
            self.procs = None

        for proc in procs:
            if isinstance(proc, (list, tuple)):
                self.set_starts(*proc, clear=False)
            elif proc not in self.starts:
                self.starts.append(proc)  # type: ignore
            else:
                raise ProcDependencyError(
                    f"{proc} is already a start process."
                )
        return self

    # In case people forget the "s"
    set_start = set_starts

    def _log_pipeline_info(self) -> None:
        """Print the information of the pipeline"""
        logger.info("")
        # Pipeline line and description
        log_rich_renderable(
            Panel(
                self.desc or Text(self.name.upper(), justify="center"),
                width=get_logpanel_width(),
                # padding=(0, 1),
                box=box.DOUBLE_EDGE,
                title=self.name.upper() if self.desc else None,
            ),
            "magenta",
            logger.info,
        )
        fmt = "[bold][magenta]%-16s:[/magenta][/bold] %s"
        enabled_plugins = (
            "{name} [cyan]{version}[/cyan]".format(
                name=name,
                version=(f"v{plg.version}" if plg.version else ""),
            )
            for name, plg in plugin.get_enabled_plugins().items()
            if name != "core"
        )
        for i, plug in enumerate(enabled_plugins):
            logger.info(fmt, "plugins" if i == 0 else "", plug)
        logger.info(fmt, "# procs", len(self.procs))
        logger.info(fmt, "profile", self.profile)
        logger.info(fmt, "outdir", self.outdir)
        logger.info(fmt, "cache", self.config.cache)
        logger.info(fmt, "dirsig", self.config.dirsig)
        logger.info(fmt, "error_strategy", self.config.error_strategy)
        logger.info(fmt, "forks", self.config.forks)
        logger.info(fmt, "lang", self.config.lang)
        logger.info(fmt, "loglevel", self.config.loglevel)
        logger.info(fmt, "num_retries", self.config.num_retries)
        logger.info(fmt, "scheduler", self.config.scheduler)
        logger.info(fmt, "submission_batch", self.config.submission_batch)
        logger.info(fmt, "template", self.config.template)
        logger.info(fmt, "workdir", self.workdir)
        for i, (key, val) in enumerate(self.config.plugin_opts.items()):
            logger.info(fmt, "plugin_opts" if i == 0 else "", f"{key}={val}")
        for i, (key, val) in enumerate(self.config.scheduler_opts.items()):
            logger.info(
                fmt, "scheduler_opts" if i == 0 else "", f"{key}={val}"
            )
        for i, (key, val) in enumerate(self.config.template_opts.items()):
            logger.info(fmt, "template_opts" if i == 0 else "", f"{key}={val}")

    async def _init(self) -> None:
        """Compute the configurations for the pipeline based on the priorities

        Configurations (priority from low to high)
        1. The default config in .defaults
        2. The plugin_opts defined in plugins (via on_setup() hook)
           (see __init__())
        3. Configuration files
        4. **kwargs from Pipen(..., **kwargs)
        5. Those defined in each Proc class
        """
        # Then load the configurations from config files
        config = ProfileConfig.load(
            {"default": self.config},
            *CONFIG_FILES,
            ignore_nonexist=True,
        )
        self.config = ProfileConfig.use_profile(
            config, self.profile, copy=True
        )

        # configs from files and CONFIG are loaded
        # allow plugins to change the default configs
        await plugin.hooks.on_init(self)
        self.workdir.mkdir(parents=True, exist_ok=True)
        # Then load the extra configurations passed from __init__(**kwargs)
        # Make sure dict options get inherited
        self.config.template_opts.update(self._kwargs.pop("template_opts", {}))
        self.config.scheduler_opts.update(
            self._kwargs.pop("scheduler_opts", {})
        )
        self.config.plugin_opts.update(self._kwargs.pop("plugin_opts", {}))
        self.config.update(self._kwargs)

    def build_proc_relationships(self) -> None:
        """Build the proc relationships for the pipeline"""
        if self.procs:
            return

        if not self.starts:
            raise ProcDependencyError(
                "No start processes specified. "
                "Did you forget to call `Pipen.set_starts()`?"
            )

        # build proc relationships
        # Allow starts to be set as a tuple
        self.procs = list(self.starts)
        nexts = set(
            sum((proc.nexts or [] for proc in self.procs), [])  # type: ignore
        )
        logger.debug("")
        logger.debug("Building process relationships:")
        logger.debug("- Start processes: %s", self.procs)
        while nexts:
            logger.debug("- Next processes: %s", nexts)
            # pick up one that can be added to procs
            for proc in sorted(
                nexts, key=lambda prc: (prc.order or 0, prc.name)
            ):
                if proc in self.procs:
                    raise ProcDependencyError(
                        f"Cyclic dependency: {proc.name}"
                    )

                if proc.name in [p.name for p in self.procs]:
                    raise PipenOrProcNameError(
                        f"'{proc.name}' is already used by another process."
                    )

                # Add proc to self.procs if all their requires
                # are added to self.procs
                # Then remove proc from nexts
                # If there are still procs in nexts
                # meaning some requires of those procs cannot run before
                # those procs.
                if not set(proc.requires) - set(self.procs):  # type: ignore
                    self.procs.append(proc)  # type: ignore
                    nexts.remove(proc)
                    nexts |= set(proc.nexts or ())
                    break
            else:
                if nexts:
                    raise ProcDependencyError(
                        f"No available next processes for {nexts}. "
                        "Did you forget to start with their "
                        "required processes?"
                    )

        self.pbar = PipelinePBar(len(self.procs), self.name.upper())



"""Provide Cli class"""
from __future__ import annotations

from abc import ABC, abstractmethod, abstractproperty
from typing import TYPE_CHECKING

from simplug import Simplug

from ..defaults import CLI_ENTRY_GROUP

if TYPE_CHECKING:
    from argx import ArgumentParser
    from argparse import Namespace

cli_plugin = Simplug(CLI_ENTRY_GROUP)


class CLIPlugin(ABC):
    """The abc for cli plugin"""

    def __init__(
        self,
        parser: ArgumentParser,
        subparser: ArgumentParser,
    ) -> None:
        self.parser = parser
        self.subparser = subparser

    @abstractproperty
    def name(self) -> str:
        """The name/command of this plugin"""

    def parse_args(self) -> Namespace:
        """Define arguments for the command"""
        return self.parser.parse_args()

    @abstractmethod
    def exec_command(self, args: Namespace) -> None:
        """Execute the command"""



"""Provide the Job class"""
from __future__ import annotations

import logging
import shlex
import shutil
from functools import cached_property
from os import PathLike
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Mapping

from diot import OrderedDiot
from xqute import Job as XquteJob
from xqute.utils import a_read_text

from ._job_caching import JobCaching
from .defaults import ProcInputType, ProcOutputType
from .exceptions import (
    ProcInputTypeError,
    ProcOutputNameError,
    ProcOutputTypeError,
    ProcOutputValueError,
    TemplateRenderingError,
)
from .template import Template
from .utils import logger, strsplit

if TYPE_CHECKING:  # pragma: no cover
    from .proc import Proc


class Job(XquteJob, JobCaching):
    """The job for pipen"""

    __slots__ = ("proc", "_output_types", "_outdir")

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.proc: Proc = None
        self._output_types: Dict[str, str] = {}
        self._outdir = self.metadir / "output"

    @property
    def script_file(self) -> Path:
        """Get the path to script file

        Returns:
            The path to the script file
        """
        return self.metadir / "job.script"

    @cached_property
    def outdir(self) -> Path:
        """Get the path to the output directory

        Returns:
            The path to the job output directory
        """
        ret = Path(self._outdir)
        # if ret is a dead link
        # when switching a proc from end/nonend to nonend/end
        if ret.is_symlink() and not ret.exists():
            ret.unlink()  # pragma: no cover
        ret.mkdir(parents=True, exist_ok=True)
        # If it is somewhere else, make a symbolic link to the metadir
        metaout = self.metadir / "output"
        if ret != metaout:
            if metaout.is_symlink() or metaout.is_file():
                metaout.unlink()
            elif metaout.is_dir():
                shutil.rmtree(metaout)
            metaout.symlink_to(ret)
        return ret

    @cached_property
    def input(self) -> Mapping[str, Any]:
        """Get the input data for this job

        Returns:
            A key-value map, where keys are the input keys
        """
        import pandas

        ret = self.proc.input.data.iloc[self.index, :].to_dict()
        # check types
        for inkey, intype in self.proc.input.type.items():

            if intype == ProcInputType.VAR or ret[inkey] is None:
                continue  # pragma: no cover, covered actually

            if intype in (ProcInputType.FILE, ProcInputType.DIR):
                if not isinstance(ret[inkey], (str, PathLike)):
                    raise ProcInputTypeError(
                        f"Got {type(ret[inkey])} instead of PathLike object "
                        f"for input: {inkey + ':' + intype!r}"
                    )

                # we should use it as a string
                ret[inkey] = str(Path(ret[inkey]).resolve())

            if intype in (ProcInputType.FILES, ProcInputType.DIRS):
                if isinstance(ret[inkey], pandas.DataFrame):
                    # // todo: nested dataframe
                    ret[inkey] = ret[inkey].iloc[0, 0]  # pragma: no cover

                if not isinstance(ret[inkey], (list, tuple)):
                    raise ProcInputTypeError(
                        f"[{self.proc.name}] Expected a sequence for input: "
                        f"{inkey + ':' + intype!r}, got {type(ret[inkey])}"
                    )

                for i, file in enumerate(ret[inkey]):
                    ret[inkey][i] = str(Path(file).resolve())
        return ret

    @cached_property
    def output(self) -> Mapping[str, Any]:
        """Get the output data of the job

        Returns:
            The key-value map where the keys are the output keys
        """
        output_template = self.proc.output
        if not output_template:
            return {}

        data = {
            "job": dict(
                index=self.index,
                metadir=str(self.metadir),
                outdir=str(self.outdir),
                stdout_file=str(self.stdout_file),
                stderr_file=str(self.stderr_file),
                jid_file=str(self.jid_file),
            ),
            "in": self.input,
            "in_": self.input,
            "proc": self.proc,
            "envs": self.proc.envs,
        }
        try:
            if isinstance(output_template, Template):
                # // TODO: check ',' in output value?
                outputs = strsplit(output_template.render(data), ",")
            else:
                outputs = [oput.render(data) for oput in output_template]
        except Exception as exc:
            raise TemplateRenderingError(
                f"[{self.proc.name}] Failed to render output."
            ) from exc

        ret = OrderedDiot()
        for oput in outputs:
            if ":" not in oput:
                raise ProcOutputNameError(
                    f"[{self.proc.name}] No name given in output."
                )

            if oput.count(":") == 1:
                output_name, output_value = oput.split(":")
                output_type = ProcOutputType.VAR
            else:
                output_name, output_type, output_value = oput.split(":", 2)
                if output_type not in ProcOutputType.__dict__.values():
                    raise ProcOutputTypeError(
                        f"[{self.proc.name}] "
                        f"Unsupported output type: {output_type}"
                    )

            self._output_types[output_name] = output_type
            ret[output_name] = output_value

            if output_type == ProcOutputType.VAR:
                continue

            if Path(output_value).is_absolute():
                raise ProcOutputValueError(
                    f"[{self.proc.name}] Path in output should be relative."
                )

            ret[output_name] = self.outdir.resolve() / output_value

            if output_type == ProcOutputType.DIR:
                ret[output_name].mkdir(parents=True, exist_ok=True)

            ret[output_name] = str(ret[output_name])

        return ret

    @cached_property
    def template_data(self) -> Mapping[str, Any]:
        """Get the data for template rendering

        Returns:
            The data for template rendering
        """

        return {
            "job": dict(
                index=self.index,
                metadir=str(self.metadir),
                outdir=str(self.outdir),
                stdout_file=str(self.stdout_file),
                stderr_file=str(self.stderr_file),
                jid_file=str(self.jid_file),
            ),
            "in": self.input,
            "in_": self.input,
            "out": self.output,
            "proc": self.proc,
            "envs": self.proc.envs,
        }

    def log(
        self,
        level: int | str,
        msg: str,
        *args,
        limit: int = 3,
        limit_indicator: bool = True,
        logger: logging.LoggerAdapter = logger,
    ) -> None:
        """Log message for the jobs

        Args:
            level: The log level of the record
            msg: The message to log
            *args: The arguments to format the message
            limit: limitation of the log (don't log for all jobs)
            limit_indicator: Whether to show an indicator saying the log
                has been limited (the level of the indicator will be DEBUG)
            logger: The logger used to log
        """
        if self.index > limit:
            return

        if self.index == limit:
            if limit_indicator:
                msg = f"{msg} (not showing similar logs)"

        job_index_indicator = "[%s/%s] " % (
            str(self.index).zfill(len(str(self.proc.size - 1))),
            self.proc.size - 1,
        )

        self.proc.log(level, job_index_indicator + msg, *args, logger=logger)

    async def prepare(self, proc: Proc) -> None:
        """Prepare the job by given process

        Primarily prepare the script, and provide cmd to the job for xqute
        to wrap and run

        Args:
            proc: the process object
        """
        # Attach the process
        self.proc = proc

        if self.proc.export and len(self.proc.jobs) == 1:
            # Don't put index if it is a single-job process
            self._outdir = Path(self.proc.pipeline.outdir) / self.proc.name

        elif self.proc.export:
            self._outdir = (
                Path(self.proc.pipeline.outdir)
                / self.proc.name
                / str(self.index)
            )

        if not proc.script:
            self.cmd = []
            return

        template_data = self.template_data
        try:
            script = proc.script.render(template_data)
        except Exception as exc:
            raise TemplateRenderingError(
                f"[{self.proc.name}] Failed to render script."
            ) from exc
        if (
            self.script_file.is_file()
            and await a_read_text(self.script_file) != script
        ):
            self.log("debug", "Job script updated.")
            self.script_file.write_text(script)
        elif not self.script_file.is_file():
            self.script_file.write_text(script)

        lang = proc.lang or proc.pipeline.config.lang
        self.cmd = shlex.split(lang) + [self.script_file]  # type: ignore



"""List available profiles."""
from __future__ import annotations
from typing import TYPE_CHECKING

import rtoml  # type: ignore
from rich import print
from rich.panel import Panel
from rich.syntax import Syntax
from simpleconf import ProfileConfig

from ._hooks import CLIPlugin
from ..defaults import CONFIG, CONFIG_FILES

if TYPE_CHECKING:
    from argx import ArgumentParser
    from argparse import Namespace

__all__ = ("CLIProfilePlugin",)


class CLIProfilePlugin(CLIPlugin):
    """List available profiles."""

    name = "profile"

    def __init__(
        self,
        parser: ArgumentParser,
        subparser: ArgumentParser,
    ) -> None:
        super().__init__(parser, subparser)
        subparser.add_argument(
            "-n",
            "--name",
            default="",
            help="The name of the profile to show. Show all if not provided.",
        )
        subparser.add_argument(
            "-l",
            "--list",
            action="store_true",
            default=False,
            help="List the names of all available profiles (-n won't work).",
        )

    def exec_command(self, args: Namespace) -> None:
        """Run the command"""

        config = ProfileConfig.load(
            {"default": CONFIG},
            *CONFIG_FILES,
            ignore_nonexist=True,
        )

        if args.list:
            print("\n".join(ProfileConfig.profiles(config)))
            return

        print("Configurations loaded from:")
        print("- pipen.defaults.CONFIG (python dictionary)")
        for conffile in reversed(CONFIG_FILES):
            print(f"- {conffile}")
        print("")

        print("Note:")
        print(
            "- The same profile from different configuration files "
            "are inherited."
        )
        print(
            "- These configurations can still be overriden by "
            "Pipen constructor and process definition."
        )
        print("")

        if not args.name:
            for profile in ProfileConfig.profiles(config):
                with ProfileConfig.with_profile(config, profile):
                    conf = ProfileConfig.detach(config)
                    print(
                        Panel(
                            Syntax(rtoml.dumps(conf), "toml"),
                            title=f"Profile: {profile}",
                            title_align="left",
                        )
                    )

        else:
            if not ProfileConfig.has_profile(config, args.name):
                raise ValueError(f"No such profile: {args.name}")

            ProfileConfig.use_profile(config, args.name)
            conf = ProfileConfig.detach(config)
            print(
                Panel(
                    Syntax(rtoml.dumps(conf), "toml"),
                    title=f"Profile: {args.name}",
                    title_align="left",
                )
            )



"""Provide exception classes"""


class PipenException(Exception):
    """Base exception class for pipen"""


class PipenSetDataError(PipenException, ValueError):
    """When trying to set input data to processes with input_data already set
    using Pipen.set_data()."""


class ProcInputTypeError(PipenException, TypeError):
    """When an unsupported input type is provided"""


class ProcInputKeyError(PipenException, KeyError):
    """When an unsupported input value is provided"""


class ProcScriptFileNotFound(PipenException, FileNotFoundError):
    """When script file specified as 'file://' cannot be found"""


class ProcOutputNameError(PipenException, NameError):
    """When no name or malformatted output is provided"""


class ProcOutputTypeError(PipenException, TypeError):
    """When an unsupported output type is provided"""


class ProcOutputValueError(PipenException, ValueError):
    """When a malformatted output value is provided"""


class ProcDependencyError(PipenException):
    """When there is something wrong the process dependencies"""


class NoSuchSchedulerError(PipenException):
    """When specified scheduler cannot be found"""


class WrongSchedulerTypeError(PipenException, TypeError):
    """When specified scheduler is not a subclass of Scheduler"""


class NoSuchTemplateEngineError(PipenException):
    """When specified template engine cannot be found"""


class WrongTemplateEnginTypeError(PipenException, TypeError):
    """When specified tempalte engine is not a subclass of Scheduler"""


class TemplateRenderingError(PipenException):
    """Failed to render a template"""


class ConfigurationError(PipenException):
    """When something wrong set as configuration"""


class PipenOrProcNameError(PipenException):
    """ "When more than one processes are sharing the same workdir"""



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pipen', 'pipen.cli']

package_data = \
{'': ['*']}

install_requires = \
['argx>=0.2.10,<0.3.0',
 'enlighten>=1,<2',
 'liquidpy>=0.8.1,<0.9.0',
 'pandas>=2,<3',
 'pipda>=0.13,<0.14',
 'python-simpleconf[toml]>=0.6,<0.7',
 'rtoml>=0.8,<0.9',
 'varname>=0.12,<0.13',
 'xqute>=0.2.4,<0.3.0']

entry_points = \
{'console_scripts': ['pipen = pipen.cli:main']}

setup_kwargs = {
    'name': 'pipen',
    'version': '0.12.3',
    'description': 'A pipeline framework for python',
    'long_description': '<div align="center">\n    <img src="./pipen.png" width="320px">\n\n**A pipeline framework for python**\n\n</div>\n\n______________________________________________________________________\n\n[![Pypi][6]][7] [![Github][8]][9] ![Building][10] [![Docs and API][11]][1] [![Codacy][12]][13] [![Codacy coverage][14]][13] [![Deps][5]][23]\n\n[Documentation][1] | [ChangeLog][2] | [Examples][3] | [API][4]\n\n## Features\n\n- Easy to use\n- Nearly zero-configuration\n- Nice logging\n- Highly extendable\n\n## Installation\n\n```bash\npip install -U pipen\n```\n\n## Quickstart\n\n`example.py`\n\n```python\nfrom pipen import Proc, Pipen\n\nclass P1(Proc):\n    """Sort input file"""\n    input = "infile"\n    input_data = ["/tmp/data.txt"]\n    output = "outfile:file:intermediate.txt"\n    script = "cat {{in.infile}} | sort > {{out.outfile}}"\n\nclass P2(Proc):\n    """Paste line number"""\n    requires = P1\n    input = "infile"\n    output = "outfile:file:result.txt"\n    script = "paste <(seq 1 3) {{in.infile}} > {{out.outfile}}"\n\nclass MyPipeline(Pipen):\n    starts = P1\n\nif __name__ == "__main__":\n    MyPipeline().run()\n```\n\n```shell\n> echo -e "3\\n2\\n1" > /tmp/data.txt\n> python example.py\n```\n\n```log\n06-09 23:15:29 I core                  _____________________________________   __\n06-09 23:15:29 I core                  ___  __ \\___  _/__  __ \\__  ____/__  | / /\n06-09 23:15:29 I core                  __  /_/ /__  / __  /_/ /_  __/  __   |/ /\n06-09 23:15:29 I core                  _  ____/__/ /  _  ____/_  /___  _  /|  /\n06-09 23:15:29 I core                  /_/     /___/  /_/     /_____/  /_/ |_/\n06-09 23:15:29 I core\n06-09 23:15:29 I core                              version: 0.12.1\n06-09 23:15:29 I core\n06-09 23:15:29 I core    ╔═══════════════════════════════════════════════════╗\n06-09 23:15:29 I core    ║                            MYPIPELINE                            ║\n06-09 23:15:29 I core    ╚═══════════════════════════════════════════════════╝\n06-09 23:15:29 I core    plugins         : verbose v0.9.0\n06-09 23:15:29 I core    # procs         : 2\n06-09 23:15:29 I core    profile         : default\n06-09 23:15:29 I core    outdir          : /home/pwwang/github/pipen/MyPipeline-output\n06-09 23:15:29 I core    cache           : True\n06-09 23:15:29 I core    dirsig          : 1\n06-09 23:15:29 I core    error_strategy  : ignore\n06-09 23:15:29 I core    forks           : 1\n06-09 23:15:29 I core    lang            : bash\n06-09 23:15:29 I core    loglevel        : info\n06-09 23:15:29 I core    num_retries     : 3\n06-09 23:15:29 I core    scheduler       : local\n06-09 23:15:29 I core    submission_batch: 8\n06-09 23:15:29 I core    template        : liquid\n06-09 23:15:29 I core    workdir         : /home/pwwang/github/pipen/.pipen/MyPipeline\n06-09 23:15:29 I core    plugin_opts     :\n06-09 23:15:29 I core    template_opts   :\n06-09 23:15:31 I core\n06-09 23:15:31 I core    ╭──────────────────────── P1 ───────────────────────╮\n06-09 23:15:31 I core    │ Sort input file                                                  │\n06-09 23:15:31 I core    ╰──────────────────────────────────────────────────╯\n06-09 23:15:31 I core    P1: Workdir: \'/home/pwwang/github/pipen/.pipen/MyPipeline/P1\'\n06-09 23:15:31 I core    P1: <<< [START]\n06-09 23:15:31 I core    P1: >>> [\'P2\']\n06-09 23:15:31 I verbose P1: size: 1\n06-09 23:15:31 I verbose P1: [0/0] in.infile: /tmp/data.txt\n06-09 23:15:31 I verbose P1: [0/0] out.outfile:\n                 /home/pwwang/github/pipen/.pipen/MyPipeline/P1/0/output/intermediate.txt\n06-09 23:15:33 I verbose P1: Time elapsed: 00:00:02.018s\n06-09 23:15:33 I core\n06-09 23:15:33 I core    ╭════════════════════════ P2 ═══════════════════════╮\n06-09 23:15:33 I core    ║ Paste line number                                                ║\n06-09 23:15:33 I core    ╰══════════════════════════════════════════════════╯\n06-09 23:15:33 I core    P2: Workdir: \'/home/pwwang/github/pipen/.pipen/MyPipeline/P2\'\n06-09 23:15:33 I core    P2: <<< [\'P1\']\n06-09 23:15:33 I core    P2: >>> [END]\n06-09 23:15:33 I verbose P2: size: 1\n06-09 23:15:33 I verbose P2: [0/0] in.infile:\n                 /home/pwwang/github/pipen/.pipen/MyPipeline/P1/0/output/intermediate.txt\n06-09 23:15:33 I verbose P2: [0/0] out.outfile:\n                 /home/pwwang/github/pipen/MyPipeline-output/P2/result.txt\n06-09 23:15:35 I verbose P2: Time elapsed: 00:00:02.009s\n06-09 23:15:35 I core\n\n              MYPIPELINE: 100%|█████████████████████████████| 2/2 [00:06<00:00, 0.36 procs/s]\n```\n\n```shell\n> cat ./MyPipeline-output/P2/result.txt\n1       1\n2       2\n3       3\n```\n\n## Examples\n\nSee more examples at `examples/` and a more realcase example at:\nhttps://github.com/pwwang/pipen-report/tree/master/example\n\n## Plugin gallery\n\nPlugins make `pipen` even better.\n\n- [`pipen-verbose`][15]: Add verbosal information in logs for pipen.\n- [`pipen-lock`][25]: Process lock for pipen to prevent multiple runs at the same time.\n- [`pipen-report`][16]: Generate report for pipen\n- [`pipen-filters`][17]: Add a set of useful filters for pipen templates.\n- [`pipen-diagram`][18]: Draw pipeline diagrams for pipen\n- [`pipen-annotate`][26]: Use docstring to annotate pipen processes\n- [`pipen-args`][19]: Command line argument parser for pipen\n- [`pipen-dry`][20]: Dry runner for pipen pipelines\n- [`pipen-log2file`][28]: Save running logs to file for pipen\n- [`pipen-board`][27]: Visualize configuration and running of pipen pipelines on the web\n- [`pipen-runinfo`][29]: Save running information to file for pipen\n- [`pipen-cli-init`][21]: A pipen CLI plugin to create a pipen project (pipeline)\n- [`pipen-cli-run`][22]: A pipen cli plugin to run a process or a pipeline\n- [`pipen-cli-require`][24]: A pipen cli plugin check the requirements of a pipeline\n\n\n[1]: https://pwwang.github.io/pipen\n[2]: https://pwwang.github.io/pipen/CHANGELOG\n[3]: https://pwwang.github.io/pipen/examples\n[4]: https://pwwang.github.io/pipen/api/pipen\n[5]: https://img.shields.io/librariesio/release/pypi/pipen?style=flat-square\n[6]: https://img.shields.io/pypi/v/pipen?style=flat-square\n[7]: https://pypi.org/project/pipen/\n[8]: https://img.shields.io/github/v/tag/pwwang/pipen?style=flat-square\n[9]: https://github.com/pwwang/pipen\n[10]: https://img.shields.io/github/actions/workflow/status/pwwang/pipen/build.yml?style=flat-square\n[11]: https://img.shields.io/github/actions/workflow/status/pwwang/pipen/docs.yml?label=docs&style=flat-square\n[12]: https://img.shields.io/codacy/grade/cf1c6c97e5c4480386a05b42dec10c6e?style=flat-square\n[13]: https://app.codacy.com/gh/pwwang/pipen\n[14]: https://img.shields.io/codacy/coverage/cf1c6c97e5c4480386a05b42dec10c6e?style=flat-square\n[15]: https://github.com/pwwang/pipen-verbose\n[16]: https://github.com/pwwang/pipen-report\n[17]: https://github.com/pwwang/pipen-filters\n[18]: https://github.com/pwwang/pipen-diagram\n[19]: https://github.com/pwwang/pipen-args\n[20]: https://github.com/pwwang/pipen-dry\n[21]: https://github.com/pwwang/pipen-cli-init\n[22]: https://github.com/pwwang/pipen-cli-run\n[23]: https://libraries.io/github/pwwang/pipen#repository_dependencies\n[24]: https://github.com/pwwang/pipen-cli-require\n[25]: https://github.com/pwwang/pipen-lock\n[26]: https://github.com/pwwang/pipen-annotate\n[27]: https://github.com/pwwang/pipen-board\n[28]: https://github.com/pwwang/pipen-log2file\n[29]: https://github.com/pwwang/pipen-runinfo\n',
    'author': 'pwwang',
    'author_email': 'pwwang@pwwang.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/pwwang/pipen',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
